uniform sampler2DShadow ShadowMap;
uniform sampler2D Texmap;
varying vec4 ProjShadow;

uniform float	texwidth;			//texture_width
uniform int		kernel_width;		//kernel width
uniform float	kernel[9];			// kernel
uniform float	kernel_norm;		// kernel normalisation

varying vec3 L;
varying vec3 N;
varying vec3 P;

// Performs 4-sample percentage closer filtering (PCF)
float pcf (vec3 shadow_coord)
{
	float resolution = 1.0 / texwidth;	// size of a texel in texture coordinates
	
	// locate lowerleft-texel location
	vec2 texel_coord = max(shadow_coord.st * texwidth - 0.5, 0.0);
	vec3 lltexel = vec3((floor(texel_coord) + 0.5) * resolution, shadow_coord.p);
	
	// take 4 samples
	float sample1 = shadow2D(ShadowMap, lltexel).x,
	      sample2 = shadow2D(ShadowMap, lltexel + vec3(resolution, 0, 0)).x,
	      sample3 = shadow2D(ShadowMap, lltexel + vec3(0, resolution, 0)).x,
	      sample4 = shadow2D(ShadowMap, lltexel + vec3(resolution, resolution, 0)).x;
	
	// take a weighted average of them, using bilinear filtering
	vec2 factor = fract(texel_coord);
	return (1.0-factor.t) * ((1.0-factor.s) * sample1 + factor.s * sample2) +
	       factor.t       * ((1.0-factor.s) * sample3 + factor.s * sample4);
}

// Performs 9-sample percentage closer filtering (PCF)
float pcf2 (vec3 shadowST)
{

	float mapScale = 1.0 / texwidth; //match this to shadowmap size
	float shadowColor = shadow2D(ShadowMap, shadowST).x;
		
	shadowColor += shadow2D(ShadowMap, shadowST.xyz + vec3( mapScale,  mapScale, 0)).x;
	shadowColor += shadow2D(ShadowMap, shadowST.xyz + vec3( mapScale, -mapScale, 0)).x;
	shadowColor += shadow2D(ShadowMap, shadowST.xyz + vec3( mapScale,  	  0, 0)).x;
	shadowColor += shadow2D(ShadowMap, shadowST.xyz + vec3(-mapScale,  mapScale, 0)).x;
	shadowColor += shadow2D(ShadowMap, shadowST.xyz + vec3(-mapScale, -mapScale, 0)).x;
	shadowColor += shadow2D(ShadowMap, shadowST.xyz + vec3(-mapScale,  	  0, 0)).x;
	shadowColor += shadow2D(ShadowMap, shadowST.xyz + vec3(        0,  mapScale, 0)).x;
	shadowColor += shadow2D(ShadowMap, shadowST.xyz + vec3(        0, -mapScale, 0)).x;
	return shadowColor = shadowColor / 9.0; //9 samples
}

// Perfoms a 3x3 gaussian filter
float gaussian(vec3 shadow_cord)
{
	float stepsize	= 1.0/texwidth;

	//apply our kernel
	float shadowColor;
	shadowColor	= shadow2D(ShadowMap,shadow_cord.xyz+vec3(-stepsize,stepsize,0)).x  * kernel[0];
	shadowColor	+= shadow2D(ShadowMap,shadow_cord.xyz+vec3(0, stepsize,0)).x  * kernel[1];
	shadowColor	+= shadow2D(ShadowMap,shadow_cord.xyz+vec3(stepsize, stepsize,0)).x  * kernel[2];
	shadowColor	+= shadow2D(ShadowMap,shadow_cord.xyz+vec3(-stepsize,0,0)).x  * kernel[3];
	shadowColor	+= shadow2D(ShadowMap,shadow_cord.xyz).x* kernel[4];
	shadowColor	+= shadow2D(ShadowMap,shadow_cord.xyz+vec3(stepsize,0,0)).x   * kernel[5];
	shadowColor	+= shadow2D(ShadowMap,shadow_cord.xyz+vec3(-stepsize,-stepsize,0)).x * kernel[6];
	shadowColor	+= shadow2D(ShadowMap,shadow_cord.xyz+vec3(0,-stepsize,0)).x  * kernel[7];
	shadowColor	+= shadow2D(ShadowMap,shadow_cord.xyz+vec3(stepsize,-stepsize,0)).x  * kernel[8];
	shadowColor /= kernel_norm;

	return shadowColor;
}
void main (void)
{
	vec3 color = vec3(0.0);
	vec4 projshad = ProjShadow;
	
	if (abs(projshad.s) < abs(projshad.q) &&
		abs(projshad.t) < abs(projshad.q) &&
		projshad.p > 0.0)
	{
		float stepsize	= 1.0/texwidth;
		
		// Perform homogenous divide
		projshad /= projshad.q;
		
		// Map from NDC to viewport
		projshad.stp = projshad.stp * 0.5 + 0.5;

		// Lookup shadow value
		//color = pcf(projshad.xyz);
		//color = pcf2(projshad.xyz);
		color = vec3(gaussian(projshad.xyz));//shadow2DProj(ShadowMap, projshad).r;
		//color = shadow2DProj(ShadowMap, projshad).r;
	}

	vec4 texcolor	= texture2D(Texmap, gl_TexCoord[0].xy);

	vec3 E = normalize(-P);
	vec3 R = normalize(-reflect(L,N));

	vec4 Iamb  = gl_FrontLightProduct[0].ambient; 
	vec4 Idiff = gl_FrontLightProduct[0].diffuse * max(dot(N,L), 0.0);
	vec4 Ispec = gl_FrontLightProduct[0].specular * pow(max(dot(R,E),0.0), gl_FrontMaterial.shininess);

	
	vec3 finalCol = texcolor.xyz*(Iamb.xyz+Idiff.xyz*color+Ispec.xyz*color); 
	//(Iamb.xyz+Idiff.xyz+Ispec.xyz)*color.xyz;//(Iamb.xyz+Idiff.xyz+Ispec.xyz);//- 0.05*color;

	gl_FragColor = vec4(finalCol, 1.0);
}